/*******************************************************************************
 * Copyright (c) 2005 Stefan Matthias Aust.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Stefan Matthias Aust - initial implementation
 *******************************************************************************/
package org.eclipse.swt.widgets;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Insets;
import java.awt.LayoutManager;
import java.util.ArrayList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.JPanel;

import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;

public class Composite extends Scrollable {
	
	protected List<Control> children = new ArrayList<Control>();
	
	protected Layout layout;
	
	Composite() {
		super();
	}
	
	public Composite(Composite parent, int style) {
		super(parent, style);
		JPanel panel = new JPanel();
		panel.setLayout(new SWTLayoutManager());
		if (_isSet(style, SWT.BORDER)) {//HACK - subclasses should not inherit this...
			panel.setBorder(BorderFactory.createTitledBorder(""));
		}
		this.container = panel; 
		parent._add(this);
	}
	
	//HACK - a table needs to avoid the normal constructor
	Composite(Composite parent, int style, int dummy) {
		super(parent, style);
	}
	
	protected void checkSubclass () {
		/* Do nothing - Subclassing is allowed */
	}

	public void setLayout(Layout layout) {
		this.layout = layout;
	}
		
	public Point computeSize(int wHint, int hHint, boolean changed) {
		Point size;
		if (layout != null) {
			size = layout.computeSize (this, wHint, hHint, changed);
		} else {
			int width = 0;
			int height = 0;
			for (Control child : children) {
				Rectangle rect = child.getBounds();
				width = Math.max(width, rect.x + rect.width);
				height = Math.max(height, rect.y + rect.height);
			}
			if (wHint != SWT.DEFAULT) {
				width = wHint;
			}
			if (hHint != SWT.DEFAULT) {
				height = hHint;
			}
			size = new Point(width, height);
		}
		Rectangle trim = computeTrim (0, 0, size.x, size.y);
		return new Point (trim.width, trim.height);
	}

	public Control[] getChildren() {
		return children.toArray(new Control[children.size()]);
	}
	
	public void layout() {
		layout(true);
	}
	
	public void layout(boolean changed) {
		if (layout != null) {
			layout.layout(this, changed);
		}
	}
	
	// copied
	public void layout(Control[] changed) {
		checkWidget();
		if (changed == null)
			error(SWT.ERROR_INVALID_ARGUMENT);
		for (int i = 0; i < changed.length; i++) {
			Control control = changed[i];
			if (control == null)
				error(SWT.ERROR_INVALID_ARGUMENT);
			if (control.isDisposed())
				error(SWT.ERROR_INVALID_ARGUMENT);
			boolean ancestor = false;
			Composite composite = control.parent;
			while (composite != null) {
				ancestor = composite == this;
				if (ancestor)
					break;
				composite = composite.parent;
			}
			if (!ancestor)
				error(SWT.ERROR_INVALID_PARENT);
		}
		int updateCount = 0;
		Composite[] update = new Composite[16];
		for (int i = 0; i < changed.length; i++) {
			Control child = changed[i];
			Composite composite = child.parent;
			while (child != this) {
				if (composite.layout != null) {
					composite.state |= LAYOUT_NEEDED;
					if (!composite.layout.flushCache(child)) {
						composite.state |= LAYOUT_CHANGED;
					}
				}
				if (updateCount == update.length) {
					Composite[] newUpdate = new Composite[update.length + 16];
					System.arraycopy(update, 0, newUpdate, 0, update.length);
					update = newUpdate;
				}
				child = update[updateCount++] = composite;
				composite = child.parent;
			}
		}
		for (int i = updateCount - 1; i >= 0; i--) {
			update[i].updateLayout(true, false);
		}
	}

	// copied
	void updateLayout(boolean resize, boolean all) {
		if (isLayoutDeferred())
			return;
		if ((state & LAYOUT_NEEDED) != 0) {
			boolean changed = (state & LAYOUT_CHANGED) != 0;
			state &= ~(LAYOUT_NEEDED | LAYOUT_CHANGED);
			if (resize)
				setResizeChildren(false);
			layout.layout(this, changed);
			if (resize)
				setResizeChildren(true);
		}
		if (all) {
			Control[] children = _getChildren();
			for (int i = 0; i < children.length; i++) {
				children[i].updateLayout(resize, all);
			}
		}
	}

	// copied
	public boolean isLayoutDeferred() {
		checkWidget();
		return layoutCount > 0 || parent.isLayoutDeferred();
	}

	// copied
	int layoutCount = 0;

	// copied and modified
	void setResizeChildren(boolean resize) {
	}

	// adapter for updateLayout
	Control[] _getChildren() {
		return getChildren();
	}

	
	// sma's extensions
	
	void _add(Control child) {
		child.parent = this;
		children.add(child);
		_control().add(child._control());
	}
	
	void _remove(Control child) {
		_control().remove(child._control());
		children.remove(child);
		child.parent = null;
	}

	class SWTLayoutManager implements LayoutManager {
		public void addLayoutComponent(String name, Component comp) {
		}

		public void removeLayoutComponent(Component comp) {
		}

		public Dimension preferredLayoutSize(Container parent) {
			if (layout == null) {
				return _control().getSize();
			}
			Point size = layout.computeSize(Composite.this, SWT.DEFAULT, SWT.DEFAULT, true);
			Insets insets = _control().getInsets();
			return new Dimension(
					insets.left + insets.right + size.x, 
					insets.top + insets.bottom + size.y);
		}

		public Dimension minimumLayoutSize(Container parent) {
			return preferredLayoutSize(parent);
		}

		public void layoutContainer(Container parent) {
			if (layout != null) {
				layout.layout(Composite.this, true);
			}
		}
	}
}
