/*******************************************************************************
 * Copyright (c) 2005 Stefan Matthias Aust.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Stefan Matthias Aust - initial implementation
 *******************************************************************************/
package org.eclipse.swt.widgets;

import java.awt.Container;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.events.TraverseListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Drawable;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.GCData;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;

public abstract class Control extends Widget implements Drawable {
	
	protected Object layoutData;
	
	protected Composite parent;

	protected boolean menuListenerInitialized;
	protected Menu menu;
	
	Control() {
		super();
	}
	
	public Control(Composite parent, int style) {
		super(parent, style);
	}

	public void setLayoutData(Object layoutData) {
		this.layoutData = layoutData;
	}

	public Object getLayoutData() {
		return layoutData;
	}
	
	public void pack() {
		setSize(computeSize(SWT.DEFAULT, SWT.DEFAULT));
	}
	
	public Point computeSize(int widthHint, int heightHint) {
		return computeSize(widthHint, heightHint, true);
	}

	public Point computeSize(int widthHint, int heightHint, boolean flushCache) {
		Dimension d = _control().getPreferredSize();
		return new Point(
				widthHint != SWT.DEFAULT ? widthHint : d.width,
				heightHint != SWT.DEFAULT ? heightHint : d.height);
	}
	
	public Rectangle computeTrim(int x, int y, int width, int height) {
		java.awt.Insets i = _control().getInsets();
		return new Rectangle(
				x - i.left, 
				y - i.top, 
				width + i.left + i.right, 
				height + i.top + i.bottom);
	}

	public Rectangle getClientArea() {
		Rectangle r = getBounds();
		java.awt.Insets i = _control().getInsets();
		return new Rectangle(i.left, i.top, r.width - i.left - i.right, r.height - i.top - i.bottom);
	}

	
	public Rectangle getBounds() {
		java.awt.Rectangle rect = _control().getBounds();
		return new Rectangle(rect.x, rect.y, rect.width, rect.height);
	}

	public void setBounds(int x, int y, int width, int height) {
		_control().setBounds(x, y, width, height);
	}

	public void setBounds(Rectangle rectangle) {
		setBounds(rectangle.x, rectangle.y, rectangle.width, rectangle.height);
	}
	
	public Point getSize() {
		Dimension size = _control().getSize();
		return new Point(size.width, size.height);
	}

	public void setSize(Point size) {
		setSize(size.x, size.y);
	}
	
	public void setSize(int x, int y) {
		_control().setSize(x, y);
	}
	
	public void setLocation(Point location) {
		setLocation(location.x, location.y);
	}
	
	public void setLocation(int x, int y) {
		_control().setLocation(x, y);
	}

	// sma's extension ---
	
	abstract Container _control();
	
	boolean _isSet(int value, int bits) {
		return (value & bits) == bits; 
	}
	
	// end

	public int getBorderWidth() {
		return 0;
	}

	public Point getLocation() {
		java.awt.Point pt = _control().getLocation();
		return new Point(pt.x, pt.y);
	}
	
	public void moveAbove(Control control) {
		if (parent.children.remove(this)) {
			parent.children.add(parent.children.indexOf(control), this);
		}
	}
	
	// copied
	void updateLayout(boolean resize, boolean all) {
		/* Do nothing */
	}
	
	public void addPaintListener(PaintListener listener) {
		checkWidget();
		if (listener == null) {
			error(SWT.ERROR_NULL_ARGUMENT);
		}
		TypedListener typedListener = new TypedListener(listener);
		addListener(SWT.Paint, typedListener);
	}

	public void removePaintListener(PaintListener listener) {
		checkWidget();
		if (listener == null) {
			error(SWT.ERROR_NULL_ARGUMENT);
		}
		if (eventTable != null) {
			eventTable.unhook(SWT.Paint, listener);
		}
	}

	public Composite getParent() {
		return parent;
	}

	public Font getFont() {
		return new Font(_control().getFont());
	}
	
	public void setFont(Font font) {
		_control().setFont(font.getAwtFont());
	}
	
	public boolean isVisible() {
		return _control().isVisible();
	}
	
	public boolean getVisible() {
		return _control().isVisible();
	}
	
	public void setVisible(boolean visible) {
		_control().setVisible(visible);
	}
	
	public boolean isEnabled() {
		return _control().isEnabled();
	}
	
	public void setEnabled(boolean enabled) {
		_control().setEnabled(enabled);
	}
	
	public void setFocus() {
		_control().requestFocus();
	}

	public int internal_new_GC(GCData data) {
		data.g = (Graphics2D) _control().getGraphics();
		return 0;
	}
	
	public void internal_dispose_GC(int handle, GCData data) {
		data.g.dispose();
	}

	protected void initMenuListener() {
		if (menuListenerInitialized) {
			return;
		}
		_realControl().addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				if (e.isPopupTrigger()) {
					popup(e);
				}
			}
		
			@Override
			public void mouseReleased(MouseEvent e) {
				if (e.isPopupTrigger()) {
					popup(e);
				}
			}
			
			private void popup(MouseEvent e) {
				((JPopupMenu) menu.component).show(e.getComponent(), e.getX(), e.getY());
			}
		});
		menuListenerInitialized = true;
	}
	
	Container _realControl() {
		Container c = _control();
		if (c instanceof JScrollPane) {
			c = (Container) ((JScrollPane) c).getViewport().getComponent(0);
		}
		return c;
	}

	public void setMenu(Menu menu) {
		initMenuListener();
		this.menu = menu;
	}

	public void setBackground(Color background) {
		_control().setBackground(background == null ? null : background.awtColor());
	}

	public void setForeground(Color foreground) {
		_control().setForeground(foreground == null ? null : foreground.awtColor());
	}

	public Shell menuShell() {
		return parent.menuShell();
	}

	public boolean isFocusControl() {
		// TODO Auto-generated method stub
		return false;
	}
	
	public Point toControl(Point point) {
		// TODO Auto-generated method stub
		return point;
	}
	
	public void addTraverseListener(TraverseListener listener) {
		checkWidget();
		if (listener == null) {
			error(SWT.ERROR_NULL_ARGUMENT);
		}
		TypedListener typedListener = new TypedListener(listener);
		addListener(SWT.Traverse, typedListener);
	}
	
	public void removeTraverseListener(TraverseListener listener) {
		checkWidget();
		if (listener == null) {
			error(SWT.ERROR_NULL_ARGUMENT);
		}
		if (eventTable != null) {
			eventTable.unhook(SWT.Traverse, listener);
		}
	}
	
	public Shell getShell() {
		return parent.getShell();
	}
	
	void releaseChild () {
		if (parent != null) parent._remove(this);
	}

}
