/*******************************************************************************
 * Copyright (c) 2005 Stefan Matthias Aust.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Stefan Matthias Aust - initial implementation
 *******************************************************************************/
package org.eclipse.swt.widgets;

import java.awt.EventQueue;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Cursor;
import org.eclipse.swt.graphics.Device;
import org.eclipse.swt.graphics.Rectangle;

public class Display extends Device {
	
	// the SWT ui thread (must be a different thread than the Swing event dispatch thread)
	final Thread thread = Thread.currentThread();
	
	// transports SWT events from Swing event dispatch thread to SWT ui thread
	final LinkedList<Event> eventQueue = new LinkedList<Event>(); 
	
	// contains all open windows to dispose them on Display#dispose()
	final List<Shell> shells = new ArrayList<Shell>();
	
	// associates threads and displays
	final static ThreadLocal<Display> displays = new ThreadLocal<Display>(); 
	
	public Display() {
		displays.set(this);
	}
	
	/**
	 * Returns the display for the current thread or <code>null</code>. 
	 */
	public static synchronized Display getCurrent() {
		return displays.get();
	}
	
	/**
	 * Sends the next pending event (sent from the Swing event dispatch thread).
	 */
	public boolean readAndDispatch() {
		if (EventQueue.isDispatchThread()) { //TODO change
			throw new Error("thread mismatch");
		}
		synchronized (eventQueue) {
			if (eventQueue.isEmpty()) {
				return false;
			}
			_sendEvent(eventQueue.removeFirst());
			return true;
		}
	}

	/**
	 * Waits until the next event is sent (from the Swing event dispatch thread). 
	 */
	public boolean sleep() {
		if (EventQueue.isDispatchThread()) { //TODO change
			throw new Error("thread mismatch");
		}
		synchronized (eventQueue) {
			try {
				eventQueue.wait();
			} catch (InterruptedException e) {
				return false;
			}
			return true;
		}
	}
	
	public void wake() {
		if (Thread.currentThread() == thread) {
			return;
		}
		synchronized (eventQueue) {
			eventQueue.notify();
		}
	}

	public void dispose() {
		for (Shell shell : shells) {
			shell.dispose();
		}
		shells.clear();
		super.dispose();
	}

	/**
	 *  
	 */
	class RunEvent extends Event {
		boolean done;
		
		RunEvent(Runnable runnable) {
			this.type = SWT.None;
			this.data = runnable;
		}
		
		void run() {
			((Runnable) data).run();
			synchronized (this) {
				done = true;
				notify();
			}
		}
	}

	
	public void asyncExec(Runnable runnable) {
		synchronized (eventQueue) {
			eventQueue.addLast(new RunEvent(runnable));
			eventQueue.notify();
		}
	}

	public void syncExec(Runnable runnable) {
		if (Thread.currentThread() == thread) {
			runnable.run();
		} else {
			RunEvent event = new RunEvent(runnable);
			synchronized (eventQueue) {
				eventQueue.addLast(event);
				eventQueue.notify();
			}
			synchronized (event) {
				try {
					while (!event.done) {
						event.wait();
					}
				} catch (InterruptedException e) {
				}
			}
		}
	}
	
	public Shell[] getShells() {
		return shells.toArray(new Shell[shells.size()]);
	}
	
	public Cursor getSystemCursor(int id) {
		return new Cursor(this, id);
	}





	// a bunch of stuff copied to make Widget work ----------
	
	/* Package Name */
	static final String PACKAGE_PREFIX = "org.eclipse.swt.widgets."; //$NON-NLS-1$
	
	boolean filters(int eventType) {
		// TODO Auto-generated method stub
		return false;
	}

	static boolean isValidClass(Class<? extends Widget> clazz) {
		String name = clazz.getName();
		int index = name.lastIndexOf('.');
		return name.substring(0, index + 1).equals(PACKAGE_PREFIX);
	}

	boolean isValidThread() {
		//return Thread.currentThread() == thread;
		return true;
	}

	boolean filterEvent(Event event) {
		// TODO Auto-generated method stub
		return false;
	}

	int getLastEventTime() {
		// TODO Auto-generated method stub
		return 0;
	}

	void postEvent(Event event) {
		// TODO Auto-generated method stub
	}
	
	// ----------
	
	/**
	 * Called from Widget to make sure that events are sent in the SWT ui thread.
	 */
	void _sendEvent(final Event event) {
		if (Thread.currentThread() == thread) {
			if (event instanceof RunEvent) {
				((RunEvent) event).run();
			} else {
				event.widget._sendEvent(event);
			}
		} else {
			synchronized (eventQueue) {
				eventQueue.addLast(event);
				eventQueue.notify();
			}
//			syncExec(new Runnable() {
//				public void run() {
//					event.widget._sendEvent(event);
//				}
//			});
		}
	}
	
	void swingCall(Runnable runnable) {
		if (EventQueue.isDispatchThread()) {
			runnable.run();
		} else {
			try {
				EventQueue.invokeAndWait(runnable);
			} catch (Exception e) {
				throw new Error(e);
			}
		}
	}

	public Rectangle getBounds() {
		return new Rectangle(0, 0, 1280, 1024);
	}

	public Thread getThread() {
		return thread;
	}

}
