/*******************************************************************************
 * Copyright (c) 2005 Stefan Matthias Aust.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Stefan Matthias Aust - initial implementation
 *******************************************************************************/
package org.eclipse.swt.widgets;

import java.awt.Container;
import java.awt.Cursor;
import java.awt.Graphics;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;

import javax.swing.BorderFactory;
import javax.swing.JPanel;
import javax.swing.border.BevelBorder;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;

public class Sash extends Control {
	
	JPanel sash;
	int startX;
	int startY;
	int lastX;
	int lastY;
	boolean dragging;

	public Sash(Composite parent, int style) {
		super(parent, style);
		sash = new JPanel();
		if (_isSet(style, SWT.BORDER)) {
			sash.setBorder(BorderFactory.createBevelBorder(BevelBorder.RAISED));
		}
		sash.setCursor(Cursor
				.getPredefinedCursor((style & SWT.VERTICAL) != 0 ? Cursor.E_RESIZE_CURSOR
						: Cursor.S_RESIZE_CURSOR)); // use same cursors as Swing
		sash.addKeyListener(new KeyAdapter() {
			@Override
			public void keyPressed(KeyEvent e) {
				switch (e.getKeyCode()) {
				case KeyEvent.VK_LEFT:
				case KeyEvent.VK_RIGHT:
				case KeyEvent.VK_UP:
				case KeyEvent.VK_DOWN:
					System.out.println(e); //TODO compute new position
					break;
				}
			}
		});
		sash.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				startX = e.getX();
				startY = e.getY();
				lastX = sash.getX();
				lastY = sash.getY();
				int width = sash.getWidth();
				int height = sash.getHeight();
				Event event = new Event();
				event.x = lastX;
				event.y = lastY;
				event.width = width;
				event.height = height;
				if ((Sash.this.style & SWT.SMOOTH) == 0) {
					event.detail = SWT.DRAG;
				}
				sendEvent(SWT.Selection, event);
				if (!isDisposed ()) {
					if (event.doit) {
						dragging = true;
						lastX = event.x;
						lastY = event.y;
						//menuShell ().bringToTop ();
						if (!isDisposed ()) {
							drawBand(event.x, event.y, width, height);
							if ((Sash.this.style & SWT.SMOOTH) != 0) {
								setBounds(event.x, event.y, width, height);
							}
						}
					}
				}
			}

			@Override
			public void mouseReleased(MouseEvent e) {
				if (dragging) {
					dragging = false;
					int width = sash.getWidth();
					int height = sash.getHeight();
					Event event = new Event();
					event.x = lastX;
					event.y = lastY;
					event.width = sash.getWidth();
					event.height = sash.getHeight();
					//drawBand(event.x, event.y, width, height);
					sash.getParent().repaint(event.x, event.y, width, height);
					sendEvent(SWT.Selection, event);
					if (!isDisposed()) {
						if (event.doit) {
							if ((Sash.this.style & SWT.SMOOTH) != 0) {
								setBounds(event.x, event.y, width, height);
							}
						}
					}
				}
			}
		});
		sash.addMouseMotionListener(new MouseMotionAdapter() {
			@Override
			public void mouseDragged(MouseEvent e) {
				int newX = e.getX() + sash.getX();
				int newY = e.getY() + sash.getY();
				int width = sash.getWidth();
				int height = sash.getHeight();
				Rectangle rect = getParent().getClientArea();
				if ((Sash.this.style & SWT.VERTICAL) != 0) {
					newX = Math.min(Math.max(0, newX - startX), rect.width - width);
					newY = lastY;
				} else {
					newX = lastX;
					newY = Math.min(Math.max(0, newY - startY), rect.height - height);
				}
				if (newX == lastX && newY == lastY) {
					return;
				}
				drawBand(lastX, lastY, width, height);
				//sash.getParent().repaint(lastX, lastX, width, height);
				Event event = new Event();
				event.x = newX;
				event.y = newY;
				event.width = width;
				event.height = height;
				if ((Sash.this.style & SWT.SMOOTH) == 0) {
					event.detail = SWT.DRAG;
				}
				sendEvent(SWT.Selection, event);
				if (!isDisposed()) { 
					if (event.doit) {
						lastX = event.x;
						lastY = event.y;
					}
					drawBand(lastX, lastY, width, height);
					if ((Sash.this.style & SWT.SMOOTH) != 0) {
						setBounds(lastX, lastY, width, height);
					}
				}
			}
		});
		parent._add(this);
	}
	
	private void drawBand(int x, int y, int width, int height) {
		if ((style & SWT.SMOOTH) == 0) {
			Graphics g = sash.getParent().getGraphics();
			g.setXORMode(sash.getParent().getBackground());
			g.drawRect(x, y, width - 1, height - 1);
			g.dispose();
		}
	}
	
	public void addSelectionListener(SelectionListener listener) {
		checkWidget();
		if (listener == null) {
			error(SWT.ERROR_NULL_ARGUMENT);
		}
		TypedListener typedListener = new TypedListener(listener);
		addListener(SWT.Selection, typedListener);
		addListener(SWT.DefaultSelection, typedListener);
	}
	
	public void removeSelectionListener(SelectionListener listener) {
		checkWidget();
		if (listener == null) {
			error(SWT.ERROR_NULL_ARGUMENT);
		}
		if (eventTable == null) {
			return;
		}
		eventTable.unhook(SWT.Selection, listener);
		eventTable.unhook(SWT.DefaultSelection, listener);
	}
	
	public Point computeSize(int wHint, int hHint, boolean changed) {
		checkWidget();
		int border = getBorderWidth();
		int width = border * 2, height = border * 2;
		if ((style & SWT.HORIZONTAL) != 0) {
			width += DEFAULT_WIDTH;
			height += 3;
		} else {
			width += 3;
			height += DEFAULT_HEIGHT;
		}
		if (wHint != SWT.DEFAULT) {
			width = wHint + border * 2;
		}
		if (hHint != SWT.DEFAULT) {
			height = hHint + border * 2;
		}
		return new Point(width, height);
	}

	@Override
	Container _control() {
		return sash;
	}

}
