/*******************************************************************************
 * Copyright (c) 2005 Stefan Matthias Aust.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Stefan Matthias Aust - initial implementation
 *******************************************************************************/
package org.eclipse.swt.widgets;

import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Window;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JMenuBar;
import javax.swing.JPanel;
import javax.swing.JRootPane;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ControlListener;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.events.ShellListener;
import org.eclipse.swt.graphics.Cursor;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Rectangle;

public class Shell extends Composite {
	
	_Window window;
	Menu menuBar;
	
	/**
	 * Constructs a new top-level shell with default window styles on the default display.
	 */
	public Shell() {
		this((Display) null, SWT.SHELL_TRIM);
	}

	/**
	 * Constructs a new top-level shell with the given style on the default display. 
	 */
	public Shell(int style) {
		this((Display) null, style);
	} 

	/**
	 * Constructs a new top-level shell with default window styles on the given display.
	 */
	public Shell(Display display) {
		this(display, SWT.SHELL_TRIM);
	}
	
	/**
	 * Constructs a new top-level shell with the given window styles on the given display.
	 */
	public Shell(Display display, int style) {
		super();
		if (display == null) {
			display = Display.getCurrent();
		}
		this.display = display;
		this.style = style;
		this.window = new _Frame(this, style);
		
		display.shells.add(this);
	}
	
	/**
	 * Constructs a new child window with default dialog styles.
	 */
	public Shell(Shell parent) {
		this(parent, SWT.DIALOG_TRIM);
	}
	
	/**
	 * Constructs a new child window with the given dialog styles.
	 */
	public Shell(Shell parent, int style) {
		super();
		this.display = parent.display;
		this.style = style;
		this.window = parent.window.child(this, style);
		
		display.shells.add(this);
	}
	
	public void open() {
		window.open();
	}
	
	public void dispose() {
		window.dispose();
		display.shells.remove(this);
		super.dispose();
	}

	public void setDefaultButton(Button button) {
		JRootPane rootPane = window.getRootPane();
		rootPane.setDefaultButton((JButton) button._control());
	}
	
	// copied
	public boolean isLayoutDeferred() {
		checkWidget();
		return layoutCount > 0;
	}

	@Override
	public Rectangle getClientArea() {
		Rectangle r = super.getClientArea();
		if (menuBar != null) {
			r.height -= window.getRootPane().getJMenuBar().getHeight();
		}
		r.x = 0;
		r.y = 0;
		return r;
	}

	public String getText() {
		return window.getText();
	}

	public void setText(String text) {
		window.setText(text);
	}

	// TODO move up
	public void addControlListener(ControlListener listener) {
		checkWidget();
		if (listener == null) {
			error(SWT.ERROR_NULL_ARGUMENT);
		}
		TypedListener typedListener = new TypedListener(listener);
		addListener(SWT.Resize, typedListener);
		addListener(SWT.Move, typedListener);
	}
	
	// TODO move up
	void fireControlMoved() {
		sendEvent(SWT.Move);
	}
	
	// TODO move up
	void fireControlResized() {
		sendEvent(SWT.Resize);
	}

	// TODO move up
	public void addPaintListener(PaintListener listener) {
		checkWidget();
		if (listener == null) {
			error(SWT.ERROR_NULL_ARGUMENT);
		}
		TypedListener typedListener = new TypedListener(listener);
		addListener(SWT.Paint, typedListener);
	}
	
	// TODO move up
	void firePaint(Graphics g) {
		Event e = new Event();
		e.gc = new GC((Graphics2D) g);
		sendEvent(SWT.Paint, e);
	}

	public void close() {
		dispose();
	}

	public void addShellListener(ShellListener listener) {
		checkWidget();
		if (listener == null) {
			error(SWT.ERROR_NULL_ARGUMENT);
		}
		TypedListener typedListener = new TypedListener(listener);
		addListener(SWT.Close, typedListener);
		addListener(SWT.Iconify, typedListener);
		addListener(SWT.Deiconify, typedListener);
		addListener(SWT.Activate, typedListener);
		addListener(SWT.Deactivate, typedListener);
	}
	
	public Menu getMenuBar() {
		return menuBar;
	}

	public void setMenuBar(Menu menuBar) {
		this.menuBar = menuBar;
		window.setMenuBar(menuBar);
		layout();
	}

	public void setCursor(final Cursor cursor) {
		display.swingCall(new Runnable() {
			public void run() {
				Component glassPane = window.getRootPane().getGlassPane();
				glassPane.setVisible(cursor != null);
				glassPane.setCursor(cursor == null ? null : cursor.awtCursor());
			}
		});
	}
	
	boolean fireClose() {
		Event event = new Event();
		sendEvent(SWT.Close, event);
		return event.doit;
	}
	
	abstract static class _Window {
		abstract Window window();
		abstract _Dialog child(Shell shell, int style); 
		abstract String getText();
		abstract void setText(String text);
		abstract void open();
		abstract void close();
		abstract void dispose();
		abstract JRootPane getRootPane();
		abstract void setMenuBar(Menu bar);
		
		void init(final Shell shell) {
			Window window = window();
			window.setSize(600, 400); //some default size
			
			window.addWindowListener(new WindowAdapter() {
				public void windowClosing(WindowEvent e) {
					shell.display.asyncExec(new Runnable() {
						public void run() {
							// this both lines must be executed in the same task without task switch
							// because otherwise the main loops goes to sleep and the application
							// doesn't terminate... a better solution is needed...
							if (shell.fireClose()) {
								shell.dispose();
							}
						}
					});
				}
			});
			
			window.addComponentListener(new ComponentListener() {
				public void componentHidden(ComponentEvent e) {
				}
			
				public void componentShown(ComponentEvent e) {
				}
			
				public void componentMoved(ComponentEvent e) {
					shell.fireControlMoved();
				}
			
				public void componentResized(ComponentEvent e) {
					shell.fireControlResized();
				}
			});
			
			JPanel contentPane = new JPanel(shell.new SWTLayoutManager()) {
				private static final long serialVersionUID = -7274157463781301008L;	
				@Override
				protected void paintComponent(Graphics g) {
					super.paintComponent(g);
					shell.firePaint(g);
				}
			};
			getRootPane().setContentPane(contentPane);

			// find a nice window location - swing insists on putting them at the same position
//			if (!display.shells.isEmpty()) {
//				Point p = display.shells.get(display.shells.size() - 1).window.getLocationOnScreen();
//				window.setLocation(p.x + 20, p.y + 20);
//			}
			shell.container = window;
		}
		
		abstract void setMinimized(boolean minimized);
		abstract void setMaximized(boolean maximized);
	}
	
	static class _Frame extends _Window {
		JFrame frame;
		
		_Frame(Shell shell, int style) {
			frame = new JFrame();
			frame.setAlwaysOnTop((style & SWT.ON_TOP) != 0);
			frame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
			frame.setLocationByPlatform(true);
			frame.setResizable((style & SWT.RESIZE) != 0);
			frame.setUndecorated((style & SWT.SHELL_TRIM) == 0);
			if (!frame.isUndecorated()) {
				frame.addNotify();
			}
			init(shell);
		}
		
		@Override
		Window window() {
			return frame;
		}
		
		@Override
		_Dialog child(Shell shell, int style) {
			return new _Dialog(shell, new JDialog(frame), style);
		}
		
		@Override
		String getText() {
			return frame.getTitle();
		}
		
		@Override
		void setText(String text) {
			frame.setTitle(text);
		}
		
		@Override
		void open() {
			frame.setVisible(true);
			frame.requestFocus();
		}
		
		@Override
		void close() {
			frame.setVisible(false);
		}
		
		@Override
		void dispose() {
			frame.dispose();
		}
		
		@Override
		JRootPane getRootPane() {
			return frame.getRootPane();
		}
		
		@Override
		void setMenuBar(Menu bar) {
			frame.setJMenuBar((JMenuBar) bar.component);
		}

		@Override
		void setMinimized(boolean minimized) {
			frame.setExtendedState(minimized ? JFrame.ICONIFIED : JFrame.NORMAL);
		}
		
		@Override
		void setMaximized(boolean maximized) {
			frame.setExtendedState(maximized ? JFrame.MAXIMIZED_BOTH : JFrame.NORMAL);
		}
	}
	
	static class _Dialog extends _Window {
		JDialog dialog;
		
		_Dialog(Shell shell, JDialog dialog, int style) {
			this.dialog = dialog;
			dialog.setAlwaysOnTop((style & SWT.ON_TOP) != 0);
			dialog.setDefaultCloseOperation(JDialog.DO_NOTHING_ON_CLOSE);
			dialog.setLocationByPlatform(true);
			dialog.setResizable((style & SWT.RESIZE) != 0);
			dialog.setUndecorated((style & SWT.DIALOG_TRIM) == 0);
			if (!dialog.isUndecorated()) {
				dialog.addNotify();
			}
			init(shell);
		}
		
		@Override
		Window window() {
			return dialog;
		}
		
		@Override
		_Dialog child(Shell shell, int style) {
			return new _Dialog(shell, new JDialog(dialog), style);
		}
		
		String getText() {
			return dialog.getTitle();
		}
		
		void setText(String text) {
			dialog.setTitle(text);
		}
		
		@Override
		void open() {
			dialog.setVisible(true);
			dialog.requestFocus();
		}
		
		@Override
		void close() {
			dialog.setVisible(false);
		}
		
		@Override
		void dispose() {
			dialog.dispose();
		}
		
		@Override
		JRootPane getRootPane() {
			return dialog.getRootPane();
		}
		
		@Override
		void setMenuBar(Menu bar) {
			dialog.setJMenuBar((JMenuBar) bar.component);
		}
		
		@Override
		void setMinimized(boolean minimized) {
			// not supported
		}
		
		@Override
		void setMaximized(boolean maximized) {
			// not supported
		}
	}

	public void setMinimized(boolean minimized) {
		window.setMinimized(minimized);
	}

	public void setMaximized(boolean maximized) {
		window.setMaximized(maximized);
	}
	
	@Override
	public Shell menuShell() {
		return this;
	}

	public void setImage(Image image) {
		// TODO Auto-generated method stub
		
	}
	
	@Override
	public Shell getShell() {
		return this;
	}

}
