/*******************************************************************************
 * Copyright (c) 2005 Stefan Matthias Aust.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Stefan Matthias Aust - initial implementation
 *******************************************************************************/
package org.eclipse.swt.widgets;

import java.awt.Component;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.ArrayList;
import java.util.List;

import javax.swing.Icon;
import javax.swing.JLabel;
import javax.swing.JTree;
import javax.swing.event.TreeExpansionEvent;
import javax.swing.event.TreeWillExpandListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.ExpandVetoException;
import javax.swing.tree.TreePath;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.events.TreeListener;
import org.eclipse.swt.graphics.Point;

public class Tree extends Scrollable {
	
	JTree tree;

	DefaultMutableTreeNode root;

	public Tree(Composite parent, int style) {
		super(parent, style);
		tree = new JTree(new DefaultTreeModel(root = new DefaultMutableTreeNode("root")));
		tree.setRootVisible(false);
		tree.setShowsRootHandles(true);
		tree.expandRow(0);
		
		// sets a renderer that displays TreeItems with text and image 
		tree.setCellRenderer(new DefaultTreeCellRenderer() {
			private static final long serialVersionUID = 4334627015204857804L;

			@Override
			public Component getTreeCellRendererComponent(JTree tree, Object value, boolean sel,
					boolean expanded, boolean leaf, int row, boolean hasFocus) {
				value = ((DefaultMutableTreeNode) value).getUserObject();
				Icon icon = null;
				if (value instanceof TreeItem) {
					icon = ((TreeItem) value)._icon();
					value = ((TreeItem) value).getText();
				}
				JLabel label = (JLabel) super.getTreeCellRendererComponent(tree, value, sel, expanded, leaf, row, hasFocus);
				label.setIcon(icon);
				return label;
			}
		});
		
		// adds a listener to generate Collapse and Expand events
		tree.addTreeWillExpandListener(new TreeWillExpandListener() {
			public void treeWillCollapse(TreeExpansionEvent e) throws ExpandVetoException {
				Event event = new Event();
				event.item = item(e);
				sendEvent(SWT.Collapse, event);
			}
		
			public void treeWillExpand(TreeExpansionEvent e) throws ExpandVetoException {
				Event event = new Event();
				event.item = item(e);
				sendEvent(SWT.Expand, event);
			}
			
			private TreeItem item(TreeExpansionEvent e) {
				return (TreeItem) ((DefaultMutableTreeNode) e.getPath().getLastPathComponent()).getUserObject();
			}
		});
		
		// adds a listener to generate Selection and DefaultSelection events
		tree.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				 TreePath path = tree.getPathForLocation(e.getX(), e.getY());
				 if (path != null) {
					 Event event = new Event();
					 event.item = (TreeItem) ((DefaultMutableTreeNode) path.getLastPathComponent()).getUserObject();
					 sendEvent(e.getClickCount() == 2 ? SWT.DefaultSelection : SWT.Selection, event);
				 }
			}
		});
			
		_init(tree, style);
		parent._add(this);
	}

	/**
	 * Returns an array containing all tree items.
	 */
	public TreeItem[] getItems() {
		List<TreeItem> items = new ArrayList<TreeItem>(); 
		items(items, root);
		return items.toArray(new TreeItem[items.size()]);
	}
	
	private void items(List<TreeItem> items, DefaultMutableTreeNode node) {
		if (node != root) {
			items.add((TreeItem) node.getUserObject());
		}
		for (int i = 0; i < node.getChildCount(); i++) {
			items(items, (DefaultMutableTreeNode) node.getChildAt(i));
		}
	}
	
	/**
	 * Returns an array containing all selected tree items.
	 */
	public TreeItem[] getSelection() {
		TreePath[] paths = tree.getSelectionPaths();
		if (paths == null) {
			return new TreeItem[0];
		}
		TreeItem[] items = new TreeItem[paths.length];
		for (int i = 0; i < items.length; i++) {
			items[i] = (TreeItem) ((DefaultMutableTreeNode) paths[i].getLastPathComponent()).getUserObject();
		}
		return items;
	}
	
	public void setSelection(TreeItem[] items) {
		int[] rows = new int[items.length];
		for (int i = 0; i < rows.length; i++) {
			rows[i] = indexOf(items[i]);
		}
		tree.setSelectionRows(rows);
	}

	int indexOf(TreeItem item) {
		TreeItem[] items = getItems();
		for (int i = 0; i < items.length; i++) {
			if (items[i] == item) {
				return i;
			}
		}
		return -1;
	}
	
	public void addSelectionListener(SelectionListener listener) {
		checkWidget();
		if (listener == null) {
			error(SWT.ERROR_NULL_ARGUMENT);
		}
		TypedListener typedListener = new TypedListener(listener);
		addListener(SWT.Selection, typedListener);
		addListener(SWT.DefaultSelection, typedListener);
	}
	
	public void removeSelectionListener(SelectionListener listener) {
		checkWidget();
		if (listener == null) {
			error(SWT.ERROR_NULL_ARGUMENT);
		}
		if (eventTable != null) {
			eventTable.unhook(SWT.Selection, listener);
			eventTable.unhook(SWT.DefaultSelection, listener);
		}
	}
	
	public void addTreeListener(TreeListener listener) {
		checkWidget();
		if (listener == null) {
			error(SWT.ERROR_NULL_ARGUMENT);
		}
		TypedListener typedListener = new TypedListener(listener);
		addListener(SWT.Expand, typedListener);
		addListener(SWT.Collapse, typedListener);
	}

	public void removeTreeListener(TreeListener listener) {
		checkWidget();
		if (listener == null) {
			error(SWT.ERROR_NULL_ARGUMENT);
		}
		if (eventTable != null) {
			eventTable.unhook(SWT.Expand, listener);
			eventTable.unhook(SWT.Collapse, listener);
		}
	}

	public TreeItem getItem(Point point) {
		TreePath path = tree.getPathForLocation(point.x, point.y);
		if (path == null) {
			return null;
		}
		return (TreeItem) ((DefaultMutableTreeNode) path.getLastPathComponent()).getUserObject();
	}
	
}
